@@ -18,11 +18,12 @@ use super::rollsum;
 use super::sendlog;
 use super::xid::*;
 use super::xtar;
+use itertools::Itertools;
 use std::cell::{Cell, RefCell};
 use std::collections::{BTreeMap, HashMap, HashSet, VecDeque};
 use std::convert::TryInto;
 use std::fmt::Write as FmtWrite;
-use std::io::{Read, Write};
+use std::io::{Read, Seek, Write};
 use std::os::unix::ffi::OsStrExt;
 use std::os::unix::fs::MetadataExt;
 use std::os::unix::fs::PermissionsExt;
@@ -39,8 +40,8 @@ pub enum ClientError {
 }
 
 pub fn open_repository(
-    w: &mut dyn std::io::Write,
-    r: &mut dyn std::io::Read,
+    w: &mut dyn Write,
+    r: &mut dyn Read,
     open_mode: OpenMode,
 ) -> Result<(), anyhow::Error> {
     write_packet(
@@ -70,8 +71,8 @@ pub fn open_repository(
 }
 
 pub fn init_repository(
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
     storage_spec: Option<repository::StorageEngineSpec>,
 ) -> Result<(), anyhow::Error> {
     write_packet(w, &Packet::TInitRepository(storage_spec))?;
@@ -120,8 +121,8 @@ struct SendSession<'a, 'b, 'c> {
     idx_tw: Cell<Option<Box<htree::TreeWriter>>>,
     idx_size: u64,
     data_size: u64,
-    r: &'c mut dyn std::io::Read,
-    w: &'c mut dyn std::io::Write,
+    r: &'c mut dyn Read,
+    w: &'c mut dyn Write,
     scratch_buf: Vec<u8>,
     log_msg_buf: String,
 }
@@ -244,7 +245,7 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
 
     fn write_data(
         &mut self,
-        data: &mut dyn std::io::Read,
+        data: &mut dyn Read,
         on_chunk: &mut dyn FnMut(&Address, usize),
     ) -> Result<u64, anyhow::Error> {
         let mut n_written: u64 = 0;
@@ -309,7 +310,8 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
     fn send_dir(
         &mut self,
         paths: Vec<std::path::PathBuf>,
-        exclusions: Vec<glob::Pattern>,
+        exclusions: globset::GlobSet,
+        exclusion_markers: std::collections::HashSet<std::ffi::OsString>,
     ) -> Result<(), anyhow::Error> {
         let use_stat_cache = self.ctx.use_stat_cache;
 
@@ -320,8 +322,10 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
             indexer::FsIndexerOptions {
                 one_file_system: self.ctx.one_file_system,
                 want_xattrs: self.ctx.want_xattrs,
+                want_sparseness: true,
                 want_hash: false,
                 exclusions,
+                exclusion_markers,
             },
         )?
         .background();
@@ -381,7 +385,7 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
                         )?;
                         index_ent.data_hash = cache_entry.hashes[i];
                         index_ent.data_cursor = cache_entry.data_cursors[i];
-                        self.write_idx_ent(&index::VersionedIndexEntry::V3(index_ent))?;
+                        self.write_idx_ent(&index::VersionedIndexEntry::V4(index_ent))?;
                     }
                 }
                 None => {
@@ -447,7 +451,6 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
                             let file_len = self.write_data(&mut f, &mut on_data_chunk)?;
 
                             if file_len != index_ent.size.0 {
-                                // Don't cache a smeared entry that is immediately invalidated by ctime.
                                 smear_detected = true;
                                 // The true size is what we read from disk.
                                 index_ent.size.0 = file_len;
@@ -486,7 +489,7 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
                             content_hashes.push(index_ent.data_hash)
                         }
 
-                        self.write_idx_ent(&index::VersionedIndexEntry::V3(index_ent))?;
+                        self.write_idx_ent(&index::VersionedIndexEntry::V4(index_ent))?;
                     }
 
                     if self.send_log_session.is_some() && use_stat_cache && !smear_detected {
@@ -512,9 +515,9 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
     }
 
     fn sync(&mut self) -> Result<(), anyhow::Error> {
-        write_packet(self.w, &Packet::TSendSync)?;
+        write_packet(self.w, &Packet::TFlush)?;
         match read_packet(self.r, DEFAULT_MAX_PACKET_SIZE)? {
-            Packet::RSendSync(stats) => {
+            Packet::RFlush(stats) => {
                 self.added_bytes += stats.added_bytes;
                 self.added_chunks += stats.added_chunks;
             }
@@ -596,11 +599,12 @@ pub enum DataSource {
     Subprocess(Vec<String>),
     Readable {
         description: String,
-        data: Box<dyn std::io::Read>,
+        data: Box<dyn Read>,
     },
     Filesystem {
         paths: Vec<std::path::PathBuf>,
-        exclusions: Vec<glob::Pattern>,
+        exclusions: globset::GlobSet,
+        exclusion_markers: std::collections::HashSet<std::ffi::OsString>,
     },
 }
 
@@ -627,8 +631,8 @@ pub struct SendStats {
 
 pub fn send(
     mut ctx: SendContext,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
     mut send_log: Option<sendlog::SendLog>,
     tags: BTreeMap<String, String>,
     data: DataSource,
@@ -709,8 +713,12 @@ pub fn send(
             ctx.progress.set_message(description);
             session.write_data(&mut data, &mut |_: &Address, _: usize| {})?;
         }
-        DataSource::Filesystem { paths, exclusions } => {
-            session.send_dir(paths, exclusions)?;
+        DataSource::Filesystem {
+            paths,
+            exclusions,
+            exclusion_markers,
+        } => {
+            session.send_dir(paths, exclusions, exclusion_markers)?;
         }
     }
 
@@ -760,8 +768,8 @@ pub fn send(
 
 pub fn request_metadata(
     id: Xid,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
 ) -> Result<oplog::VersionedItemMetadata, anyhow::Error> {
     write_packet(w, &Packet::TRequestMetadata(TRequestMetadata { id }))?;
 
@@ -788,9 +796,9 @@ pub fn request_data_stream(
     metadata: &oplog::VersionedItemMetadata,
     data_map: Option<index::DataMap>,
     index: Option<index::CompressedIndex>,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
-    out: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
+    out: &mut dyn Write,
 ) -> Result<(), anyhow::Error> {
     // It makes little sense to ask the server for an empty pick.
     if let Some(ref data_map) = data_map {
@@ -873,8 +881,8 @@ pub fn request_index(
     mut ctx: IndexRequestContext,
     id: Xid,
     metadata: &oplog::VersionedItemMetadata,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
 ) -> Result<index::CompressedIndex, anyhow::Error> {
     if ctx.primary_key_id != *metadata.primary_key_id() {
         anyhow::bail!("decryption key does not match key used for encryption");
@@ -922,7 +930,7 @@ pub fn request_index(
 }
 
 fn receive_and_authenticate_htree_chunk(
-    r: &mut dyn std::io::Read,
+    r: &mut dyn Read,
     address: Address,
 ) -> Result<Vec<u8>, anyhow::Error> {
     let data = match read_packet(r, DEFAULT_MAX_PACKET_SIZE)? {
@@ -944,9 +952,9 @@ fn receive_and_authenticate_htree_chunk(
 fn receive_htree(
     dctx: &mut crypto::DecryptionContext,
     hash_key: &crypto::HashKey,
-    r: &mut dyn std::io::Read,
+    r: &mut dyn Read,
     tr: &mut htree::TreeReader,
-    out: &mut dyn std::io::Write,
+    out: &mut dyn Write,
 ) -> Result<u64, anyhow::Error> {
     let mut n_copied: u64 = 0;
     while let Some((height, addr)) = tr.next_addr() {
@@ -980,12 +988,12 @@ fn receive_htree(
 fn write_indexed_data_as_tarball(
     read_data: &mut dyn FnMut() -> Result<Option<Vec<u8>>, anyhow::Error>,
     index: &index::CompressedIndex,
-    out: &mut dyn std::io::Write,
+    out: &mut dyn Write,
 ) -> Result<(), anyhow::Error> {
     let mut buffered = vec![];
     let mut buffer_index: usize = 0;
     let mut copy_out_and_hash =
-        |out: &mut dyn std::io::Write, mut n: u64| -> Result<blake3::Hash, anyhow::Error> {
+        |out: &mut dyn Write, mut n: u64| -> Result<blake3::Hash, anyhow::Error> {
             let mut hasher = blake3::Hasher::new();
 
             'read: while n != 0 {
@@ -1023,7 +1031,7 @@ fn write_indexed_data_as_tarball(
             Ok(hasher.finalize())
         };
 
-    let mut hardlinks: std::collections::HashMap<(u64, u64), String> =
+    let mut hardlinks: std::collections::HashMap<(u64, u64), PathBuf> =
         std::collections::HashMap::new();
 
     for ent in index.iter() {
@@ -1058,7 +1066,7 @@ fn write_indexed_data_as_tarball(
                 index::ContentCryptoHash::None => (), // XXX we don't currently require this, but we should for 1.0.
                 index::ContentCryptoHash::Blake3(expected_hash) => {
                     if hash_bytes != expected_hash {
-                        anyhow::bail!("entry {} content hash differs from index hash, possible corruption detected.", ent.path);
+                        anyhow::bail!("entry {} content hash differs from index hash, possible corruption detected.", ent.path.to_string_lossy());
                     }
                 }
             }
@@ -1092,10 +1100,10 @@ fn write_indexed_data_as_tarball(
 fn receive_indexed_htree_as_tarball(
     dctx: &mut crypto::DecryptionContext,
     hash_key: &crypto::HashKey,
-    r: &mut dyn std::io::Read,
+    r: &mut dyn Read,
     tr: &mut htree::TreeReader,
     index: &index::CompressedIndex,
-    out: &mut dyn std::io::Write,
+    out: &mut dyn Write,
 ) -> Result<(), anyhow::Error> {
     let mut read_data = || -> Result<Option<Vec<u8>>, anyhow::Error> {
         while let Some((height, addr)) = tr.next_addr() {
@@ -1131,12 +1139,12 @@ fn receive_indexed_htree_as_tarball(
 fn receive_partial_htree(
     dctx: &mut crypto::DecryptionContext,
     hash_key: &crypto::HashKey,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
     tr: &mut htree::TreeReader,
     data_map: index::DataMap,
     index: Option<index::CompressedIndex>,
-    out: &mut dyn std::io::Write,
+    out: &mut dyn Write,
 ) -> Result<(), anyhow::Error> {
     // This is avoided before we make a server request.
     assert!(!data_map.data_chunk_ranges.is_empty());
@@ -1263,8 +1271,8 @@ fn receive_partial_htree(
 
 pub fn recover_removed(
     progress: indicatif::ProgressBar,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
 ) -> Result<u64, anyhow::Error> {
     progress.set_message("recovering items...");
 
@@ -1277,18 +1285,15 @@ pub fn recover_removed(
 
 pub fn gc(
     progress: indicatif::ProgressBar,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
 ) -> Result<repository::GcStats, anyhow::Error> {
     progress.set_message("collecting garbage...");
 
     write_packet(w, &Packet::TGc(TGc {}))?;
     loop {
         match read_packet(r, DEFAULT_MAX_PACKET_SIZE)? {
-            Packet::Progress(Progress::Notice(msg)) => {
-                progress.println(&msg);
-            }
-            Packet::Progress(Progress::SetMessage(msg)) => {
+            Packet::GcProgress(msg) => {
                 progress.set_message(msg);
             }
             Packet::RGc(rgc) => return Ok(rgc.stats),
@@ -1300,8 +1305,8 @@ pub fn gc(
 pub fn sync_query_cache(
     progress: indicatif::ProgressBar,
     query_cache: &mut querycache::QueryCache,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
 ) -> Result<(), anyhow::Error> {
     progress.set_message("fetching remote metadata...");
 
@@ -1312,18 +1317,18 @@ pub fn sync_query_cache(
 
     write_packet(
         w,
-        &Packet::TRequestItemSync(TRequestItemSync {
+        &Packet::TRequestOpLogSync(TRequestOpLogSync {
             after: after.map(serde_bare::Uint),
             gc_generation,
         }),
     )?;
 
     let gc_generation = match read_packet(r, DEFAULT_MAX_PACKET_SIZE)? {
-        Packet::RRequestItemSync(ack) => ack.gc_generation,
+        Packet::RRequestOpLogSync(ack) => ack.gc_generation,
         _ => anyhow::bail!("protocol error, expected items packet"),
     };
 
-    tx.start_sync(gc_generation)?;
+    tx.start_oplog_sync(gc_generation)?;
 
     loop {
         match read_packet(r, DEFAULT_MAX_PACKET_SIZE)? {
@@ -1346,8 +1351,8 @@ pub fn sync_query_cache(
 pub fn remove(
     progress: indicatif::ProgressBar,
     ids: Vec<Xid>,
-    r: &mut dyn std::io::Read,
-    w: &mut dyn std::io::Write,
+    r: &mut dyn Read,
+    w: &mut dyn Write,
 ) -> Result<u64, anyhow::Error> {
     progress.set_message("removing items...");
 
@@ -1377,9 +1382,9 @@ pub fn restore_to_local_dir(
     progress: &indicatif::ProgressBar,
     ctx: RestoreContext,
     content_index: index::CompressedIndex,
-    pick: Option<String>,
-    serve_out: &mut dyn std::io::Read,
-    serve_in: &mut dyn std::io::Write,
+    pick: Option<PathBuf>,
+    serve_out: &mut dyn Read,
+    serve_in: &mut dyn Write,
     to_dir: &Path,
 ) -> Result<(), anyhow::Error> {
     let sub_index = match pick {
@@ -1441,7 +1446,9 @@ pub fn restore_to_local_dir(
         for indexed_dir in indexer::FsIndexer::new(
             &[to_dir.to_owned()],
             indexer::FsIndexerOptions {
-                exclusions: vec![],
+                exclusions: globset::GlobSet::empty(),
+                exclusion_markers: HashSet::new(),
+                want_sparseness: false,
                 want_xattrs: false,
                 want_hash: true,
                 one_file_system: false,
@@ -1464,10 +1471,10 @@ pub fn restore_to_local_dir(
     let mut downloads = Vec::with_capacity(512);
 
     {
-        let download_path_prefix = match pick {
-            Some(ref pick) if pick == "." => "".to_string(),
-            Some(ref pick) => format!("{}/", pick),
-            None => "".to_string(),
+        let download_path_prefix: PathBuf = match pick {
+            Some(ref pick) if pick == Path::new(".") => "".into(),
+            Some(ref pick) => fsutil::path_raw_join(pick, Path::new("/")),
+            None => "".into(),
         };
 
         index::diff(
@@ -1487,10 +1494,9 @@ pub fn restore_to_local_dir(
                         if e.is_dir() {
                             new_dirs.push(PathBuf::from(&e.path));
                         } else if e.is_file() {
-                            // Mapping back to the path in the unpicked index.
-                            download_index_path_set
-                                .insert(format!("{}{}", download_path_prefix, e.path));
-                            downloads.push((e.path.clone(), e.size.0, e.data_hash));
+                            let mapped_path = fsutil::path_raw_join(&download_path_prefix, &e.path);
+                            download_index_path_set.insert(mapped_path);
+                            downloads.push((e.path.clone(), e.size.0, e.data_hash, e.sparse));
                         } else {
                             create_path_set.insert(e.path.clone());
                             create_index_writer.add(e);
@@ -1548,32 +1554,51 @@ pub fn restore_to_local_dir(
         let to_dir = to_dir.to_owned();
         let worker = std::thread::spawn(move || -> std::io::Result<()> {
             let r = &mut r;
-            for (path, size, hash) in downloads.drain(..) {
+            for (path, size, hash, sparse) in downloads.drain(..) {
                 let mut to_create = to_dir.to_owned();
                 to_create.push(&path);
-                let mut f = std::fs::File::create(to_create)?;
+                let mut f = std::fs::OpenOptions::new()
+                    .write(true)
+                    .read(true)
+                    .create(true)
+                    .open(to_create)?;
+
+                let n_copied = if sparse {
+                    fsutil::copy_as_sparse_file(&mut r.take(size), &mut f)?
+                } else {
+                    std::io::copy(&mut r.take(size), &mut f)?
+                };
+
+                if n_copied != size {
+                    return Err(std::io::Error::new(
+                        std::io::ErrorKind::Other,
+                        format!(
+                            "content of {} is smaller than expected",
+                            path.to_string_lossy()
+                        ),
+                    ));
+                }
+
+                f.seek(std::io::SeekFrom::Start(0))?;
+
                 match hash {
                     index::ContentCryptoHash::None => {
                         return Err(std::io::Error::new(
                             std::io::ErrorKind::Other,
-                            format!("{} in content index missing hash", &path),
+                            format!("{} in content index missing hash", path.to_string_lossy()),
                         ))
                     }
                     index::ContentCryptoHash::Blake3(expected_hash) => {
-                        let mut tee = ioutil::TeeReader::new(r.take(size), blake3::Hasher::new());
-                        let n = std::io::copy(&mut tee, &mut f)?;
-                        if n != size {
-                            return Err(std::io::Error::new(
-                                std::io::ErrorKind::Other,
-                                format!("content of {} is smaller than expected", &path),
-                            ));
-                        }
-                        let (_, hasher) = tee.into_inner();
+                        let mut hasher = blake3::Hasher::new();
+                        std::io::copy(&mut f, &mut hasher)?;
                         let actual_hash: [u8; 32] = hasher.finalize().into();
                         if expected_hash != actual_hash {
                             return Err(std::io::Error::new(
                                 std::io::ErrorKind::Other,
-                                format!("content of {} did not match expected hash", &path),
+                                format!(
+                                    "content of {} did not match expected hash",
+                                    path.to_string_lossy()
+                                ),
                             ));
                         }
                     }
@@ -1617,7 +1642,7 @@ pub fn restore_to_local_dir(
             match ent.kind() {
                 index::IndexEntryKind::Symlink => {
                     if ent.link_target.is_none() {
-                        anyhow::bail!("{} is missing a link target", ent.path);
+                        anyhow::bail!("{} is missing a link target", ent.path.to_string_lossy());
                     }
                     match std::os::unix::fs::symlink(&ent.link_target.unwrap(), &to_create) {
                         Ok(_) => (),
@@ -1694,7 +1719,7 @@ pub fn restore_to_local_dir(
                         Ok(()) => (),
                         Err(err) => anyhow::bail!(
                             "failed to list remove xattr {} from {}: {}",
-                            attr,
+                            attr.to_string_lossy(),
                             to_ch.display(),
                             err
                         ),
@@ -1730,6 +1755,22 @@ pub fn restore_to_local_dir(
             };
         }
 
+        if let (Ok(tv_sec), Ok(tv_nsec)) = (ent.mtime.0.try_into(), ent.mtime_nsec.0.try_into()) {
+            match nix::sys::stat::utimensat(
+                None,
+                to_ch,
+                &nix::sys::time::TimeSpec::from(libc::timespec {
+                    tv_sec: 0,
+                    tv_nsec: libc::UTIME_OMIT,
+                }),
+                &nix::sys::time::TimeSpec::from(libc::timespec { tv_sec, tv_nsec }),
+                nix::sys::stat::UtimensatFlags::NoFollowSymlink,
+            ) {
+                Ok(_) => (),
+                Err(err) => anyhow::bail!("failed to set mtime of {}: {}", to_ch.display(), err),
+            };
+        }
+
         Ok(())
     };
 
@@ -1741,7 +1782,9 @@ pub fn restore_to_local_dir(
         index::diff(
             &to_dir_index,
             index_to_diff,
-            !(index::INDEX_COMPARE_MASK_PERMS | index::INDEX_COMPARE_MASK_XATTRS),
+            !(index::INDEX_COMPARE_MASK_PERMS
+                | index::INDEX_COMPARE_MASK_XATTRS
+                | index::INDEX_COMPARE_MASK_MTIME),
             &mut |ds: index::DiffStat, ent: &index::IndexEntry| -> Result<(), anyhow::Error> {
                 if matches!(ds, index::DiffStat::Removed) {
                     // Nothing to do, removals already processed.
@@ -1783,7 +1826,7 @@ pub fn restore_to_local_dir(
                         }
 
                         if matches!(ds, index::DiffStat::Added) {
-                            // Set the perms and xattrs of anything that changed.
+                            // Set the attributes of anything that changed.
                             let mut to_ch = to_dir.to_owned();
                             to_ch.push(&ent.path);
                             if ent.is_dir() {
@@ -1810,7 +1853,237 @@ pub fn restore_to_local_dir(
     Ok(())
 }
 
-pub fn hangup(w: &mut dyn std::io::Write) -> Result<(), anyhow::Error> {
+pub fn repo_sync(
+    progress: &indicatif::ProgressBar,
+    item_ids: Vec<Xid>,
+    ids_to_metadata: Option<HashMap<Xid, oplog::VersionedItemMetadata>>,
+    source_serve_out: &mut (dyn Read + Send),
+    source_serve_in: &mut (dyn Write + Send),
+    dest_serve_out: &mut (dyn Read + Send),
+    dest_serve_in: &mut (dyn Write + Send),
+) -> Result<(), anyhow::Error> {
+    write_packet(dest_serve_in, &Packet::TBeginItemSyncPush)?;
+
+    match read_packet(dest_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+        Packet::RBeginItemSyncPush(_) => (),
+        _ => anyhow::bail!("expected RBeginItemSyncPush packet"),
+    }
+
+    let mut missing_item_ids = Vec::with_capacity(256);
+
+    const ITEM_BATCH_SIZE: usize = DEFAULT_MAX_PACKET_SIZE / XID_SZ;
+    // Iterate in chunks that will fit the max packet size.
+    progress.set_message(format!("computing item send set 0/{}...", item_ids.len()));
+    let mut items_processed: u64 = 0;
+    for item_batch in item_ids.chunks(ITEM_BATCH_SIZE) {
+        write_item_sync_filter_items(dest_serve_in, item_batch)?;
+        let mut batch_progress = 0;
+        loop {
+            match read_packet(dest_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+                Packet::ItemSyncFilterItemsProgress(n) => {
+                    batch_progress += n.0;
+                    items_processed += n.0;
+                    progress.set_message(format!(
+                        "computing item send set {}/{}...",
+                        items_processed,
+                        item_ids.len(),
+                    ));
+                }
+                Packet::ItemSyncItems(mut missing) => {
+                    missing_item_ids.append(&mut missing);
+                    items_processed += (item_batch.len() as u64) - batch_progress;
+                    break;
+                }
+                _ => anyhow::bail!("expected RBeginRepoSyncPush packet"),
+            }
+        }
+    }
+
+    write_packet(source_serve_in, &Packet::TBeginItemSyncPull)?;
+
+    match read_packet(source_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+        Packet::RBeginItemSyncPull => (),
+        _ => anyhow::bail!("expected RBeginItemSyncPush packet"),
+    }
+
+    let all_chunks: Vec<Address> = {
+        progress.set_message("counting chunks...");
+        let mut all_chunks = HashSet::with_capacity(16 * 1024);
+        for missing_item_ids in missing_item_ids.chunks(ITEM_BATCH_SIZE) {
+            write_item_sync_request_addresses(source_serve_in, missing_item_ids)?;
+            loop {
+                let addresses = match read_packet(source_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+                    Packet::ItemSyncAddresses(addresses) => addresses,
+                    _ => anyhow::bail!("expected ItemSyncAddresses packet"),
+                };
+                for address in addresses.iter() {
+                    all_chunks.insert(*address);
+                }
+                progress.set_message(format!("counting chunks, {} found...", all_chunks.len(),));
+                if addresses.is_empty() {
+                    break;
+                }
+            }
+        }
+        all_chunks.into_iter().collect()
+    };
+
+    let mut all_missing_chunks = Vec::with_capacity(all_chunks.len() / 10);
+    {
+        let mut chunks_processed: u64 = 0;
+        progress.set_message("computing chunk send set...");
+        // Gather the missing chunks, doing the requests in large batches.
+        for chunk_batch in all_chunks.chunks(DEFAULT_MAX_PACKET_SIZE / ADDRESS_SZ) {
+            write_item_sync_filter_existing(dest_serve_in, chunk_batch)?;
+            let mut batch_progress = 0;
+            loop {
+                match read_packet(dest_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+                    Packet::ItemSyncFilterExistingProgress(n) => {
+                        batch_progress += n.0;
+                        chunks_processed += n.0;
+                        progress.set_message(format!(
+                            "computing chunk send set {}/{}...",
+                            chunks_processed,
+                            all_chunks.len(),
+                        ));
+                    }
+                    Packet::ItemSyncAddresses(mut missing) => {
+                        all_missing_chunks.append(&mut missing);
+                        chunks_processed += (chunk_batch.len() as u64) - batch_progress;
+                        break;
+                    }
+                    _ => anyhow::bail!("expected ItemSyncAddresses or Process packet"),
+                };
+            }
+        }
+    }
+
+    std::mem::drop(all_chunks);
+
+    progress.set_message("copying chunks");
+    progress.set_position(0);
+    progress.set_length(all_missing_chunks.len().try_into().unwrap());
+    progress.set_style(
+        indicatif::ProgressStyle::default_bar()
+            .template("[{elapsed_precise}] {msg} [{wide_bar}] {pos:>7}/{len:7}")
+            .progress_chars("=> "),
+    );
+
+    let n_chunks_to_copy = all_missing_chunks.len();
+    // Read chunks in a thread so we are always reading and writing chunks in parallel.
+    crossbeam_utils::thread::scope(|tscope| -> Result<(), anyhow::Error> {
+        let (chunk_tx, chunk_rx) = crossbeam_channel::bounded(0);
+        let copy_chunk_worker: crossbeam_utils::thread::ScopedJoinHandle<
+            Result<(), anyhow::Error>,
+        > = tscope.spawn(|_| {
+            let mut n_chunks_to_copy = n_chunks_to_copy;
+            let chunk_tx = chunk_tx;
+            while n_chunks_to_copy != 0 {
+                match read_packet(source_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+                    Packet::Chunk(chunk) => {
+                        chunk_tx.send(chunk)?;
+                        n_chunks_to_copy -= 1;
+                    }
+                    _ => anyhow::bail!("expected Chunk packet"),
+                };
+            }
+            Ok(())
+        });
+
+        for addresses in all_missing_chunks.chunks(DEFAULT_MAX_PACKET_SIZE / ADDRESS_SZ) {
+            write_item_sync_addresses(source_serve_in, addresses)?;
+            for _ in 0..addresses.len() {
+                match chunk_rx.recv() {
+                    Ok(chunk) => {
+                        write_chunk(dest_serve_in, &chunk.address, &chunk.data)?;
+                        progress.inc(1);
+                    }
+                    Err(_) => {
+                        copy_chunk_worker.join().unwrap()?;
+                        unreachable!();
+                    }
+                }
+            }
+        }
+
+        copy_chunk_worker.join().unwrap()?;
+        Ok(())
+    })
+    .unwrap()?;
+
+    // XXX indicatif doesn't provide a way to save and then restore the previous style,
+    // so we just use our knowledge of what it was to reset it.
+    progress.set_style(
+        indicatif::ProgressStyle::default_spinner().template("[{elapsed_precise}] {wide_msg}"),
+    );
+
+    write_packet(dest_serve_in, &Packet::TFlush)?;
+    match read_packet(dest_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+        Packet::RFlush(_) => (),
+        _ => anyhow::bail!("protocol error, expected RSentSync packet"),
+    }
+
+    let mut missing_item_metadata = Vec::with_capacity(missing_item_ids.len());
+    if let Some(ids_to_metadata) = ids_to_metadata {
+        for id in missing_item_ids.iter() {
+            match ids_to_metadata.get(id) {
+                Some(md) => missing_item_metadata.push(md.clone()),
+                None => anyhow::bail!("metadata missing for {}", id.to_string()),
+            }
+        }
+    } else {
+        progress.set_message(format!(
+            "fetching item metadata from source 0/{}...",
+            missing_item_ids.len(),
+        ));
+        for missing_item_ids in missing_item_ids.chunks(ITEM_BATCH_SIZE) {
+            write_packet(
+                source_serve_in,
+                &Packet::ItemSyncRequestMetadata(missing_item_ids.to_vec()),
+            )?;
+            while missing_item_metadata.len() != missing_item_ids.len() {
+                match read_packet(source_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+                    Packet::ItemSyncMetadata(mut md) => {
+                        progress.set_message(format!(
+                            "fetching item metadata from source {}/{}...",
+                            missing_item_metadata.len(),
+                            missing_item_ids.len(),
+                        ));
+                        missing_item_metadata.append(&mut md);
+                    }
+                    _ => anyhow::bail!("protocol error, expected ItemSyncMetadata packet"),
+                }
+            }
+        }
+    }
+
+    progress.set_message("adding new items...");
+    for packet in missing_item_ids
+        .into_iter()
+        .zip(missing_item_metadata.into_iter())
+        .chunks(128) // Small enough to avoid max packet size even with large metadata.
+        .into_iter()
+        .map(|it| Packet::ItemSyncAddItems(it.collect()))
+    {
+        write_packet(dest_serve_in, &packet)?;
+    }
+
+    write_packet(source_serve_in, &Packet::TEndItemSyncPull)?;
+    match read_packet(source_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+        Packet::REndItemSyncPull => (),
+        _ => anyhow::bail!("expected REndItemSyncPull packet"),
+    };
+
+    write_packet(dest_serve_in, &Packet::TEndItemSyncPush)?;
+    match read_packet(dest_serve_out, DEFAULT_MAX_PACKET_SIZE)? {
+        Packet::REndItemSyncPush => (),
+        _ => anyhow::bail!("expected REndItemSyncPush packet"),
+    };
+
+    Ok(())
+}
+
+pub fn hangup(w: &mut dyn Write) -> Result<(), anyhow::Error> {
     write_packet(w, &Packet::EndOfTransmission)?;
     Ok(())
 }
